
// client.cpp
// v0.84 (7.3.2002)
// (c) 2002 Andreas Krueger
//
// Please, if you change anything in this code,
// mark the end of those lines with '// *+* your name *+* ' 
// (so that I can find it) and send the code 
// to cye-code@AndreasKrueger.de
//
// last change: 
//          by:



#include"client.h"
#include"storage.h"
#include"constants.h"
#include<qmessagebox.h>
#include<qcheckbox.h>
#include<qgroupbox.h>

typedef ClientStorage myData;


Client::Client(myData * DataPointer, Dialogs * aDialogsPointer,
			   QWidget * parent, const char * name)
	: QVBox(parent, name), myParent(parent),
	  DataStoragePointer(DataPointer), DialogsPointer (aDialogsPointer)
{
	parent->setCaption(tr("CYE client (%1) for data transfer to server").arg(SCVERSION));

	QString itext = QString(
		"\nAndreasKrueger.de very proudly presents:\n"
		"The CLIENT %1 of CYE v%2 - "
		"Please connect to the Server now.\n"
		);
	itext=itext.arg(SCVERSION).arg(PROGRAMVERSION/100.0);

	QLabel *lb = new QLabel( itext, this );
	lb->setAlignment( AlignHCenter | AlignVCenter);

	if (debugging){
		QHBox *hostedit = new QHBox( this );
		editHostaddress = new QLineEdit( hostedit );
		editHostport = new QLineEdit( hostedit );
		QPushButton *setAddress = new QPushButton( tr("Set &Server Address") , hostedit );
		connect(setAddress, SIGNAL(clicked ()),	this, SLOT(setHost() ) );
	
		QCheckBox *all = new QCheckBox(tr("switch to single operation steps (for BETA only)"), this);
		all->setChecked(true);
		connect( all, SIGNAL(toggled(bool)), SLOT(TESTToggleOpenSendProcessAnswerClose(bool)) );
	}

	open = new QPushButton( tr("&Open Connection") , this );
	open->setGeometry(QRect(0,0,this->width(), 60));
	infoText = new QTextView( this );
	infoText->setTextFormat(PlainText);
	// TODO: how to give the QTextView a AUTO SCROLL behaviour?

	QPushButton *clear = new QPushButton( tr("Clear &Window") , this );
	connect( clear, SIGNAL(clicked()), SLOT(clearView()) );

	if (debugging) {
		TESTsendBox = new QGroupBox(5, Qt::Vertical, this);
		QLabel *betalabel;
 		  betalabel= new QLabel("This is for testing the beta version", TESTsendBox);
		QPushButton *send = new QPushButton( tr("Send &Data") , TESTsendBox );
		QHBox *hb = new QHBox( TESTsendBox  );
		QPushButton *TESTSendWrong = new QPushButton(tr("Send &Faulty Format"),hb);
		QPushButton *TESTSendTooMuch = new QPushButton(tr("Send &Too Much"),hb);
		QHBox *hb2 = new QHBox( TESTsendBox  );
		QPushButton *TESTSendOldVersion = new QPushButton(tr("Send Old &Version"),hb2);
		QPushButton *TESTTooNewVersion = new QPushButton( tr("Send &Too New Version"),hb2);
		QPushButton *close = new QPushButton( tr("&Close Connection") , TESTsendBox );
		TESTsendBox->setEnabled(false);

		connect( send, SIGNAL(clicked()), SLOT(sendToServer()) );
		connect( TESTSendWrong, SIGNAL(clicked()), SLOT(TESTSendWrongFormatToServer()) );
		connect( TESTSendOldVersion, SIGNAL(clicked()), SLOT(TESTSendOldVersionToServer()) );
		connect( TESTTooNewVersion, SIGNAL(clicked()), SLOT(TESTSendTooNewVersionToServer()) );
		connect( TESTSendTooMuch, SIGNAL(clicked()), SLOT(TESTSendTooMuchToServer()) );
		//	connect( TESTSlowSend, SIGNAL(clicked()), SLOT(TESTSlowSendToServer()) );
		connect( close, SIGNAL(clicked()), SLOT(closeConnection()) );
	}

	QPushButton *quit = new QPushButton( tr("&Quit") , this );

	connect( open, SIGNAL(clicked()), SLOT(openConnection()) );
	connect( quit, SIGNAL(clicked()), this, SLOT(quit()) );

	// create the socket and connect various of its signals
	theSocket = new QSocket( this );
	connect( theSocket, SIGNAL(hostFound()), SLOT(socketHostFound()) );
	connect( theSocket, SIGNAL(connected()), SLOT(socketConnected()) );
	connect( theSocket, SIGNAL(connectionClosed()), SLOT(socketConnectionClosed()) );
	connect( theSocket, SIGNAL(readyRead()), SLOT(socketReadyRead()) );
	connect( theSocket, SIGNAL(error(int)),  SLOT(socketError(int)) );

	connect(this, SIGNAL(IllegalConnection(int)), this, SLOT(CloseTheConnection(int)) );
	connect(this, SIGNAL(ClientProblem(int)), this, SLOT(explainProblem(int)) );

	DNSTimer=new QTimer(this);
	connect (DNSTimer, SIGNAL(timeout()), this, SLOT(HostNotFoundInDNS()));

	ConnectTimer=new QTimer(this);
	connect (ConnectTimer, SIGNAL(timeout()), this, SLOT(ServerNotFound()));
}

void Client::initClient(const QString &host, Q_UINT16 port){
	servername=host;
	serverport=port;
    if (debugging){
		editHostaddress->setText(servername);
		editHostport->setText(tr("%1").arg(serverport));
	}

	HeaderRead=false;  // the mirrored data is still of unknown size
	noWait=true;
	uploadSuccessful=false;
	emit Event("ClientStarted");
}

void Client::quit(){
	if (uploadSuccessful){
		emit Event("ClientUploadSuccessful");
		// not the most elegant, but it works:
		QTimer::singleShot(30, myParent, SLOT(accept()) );
	}
	else{
		emit Event("ClientUploadQuitWithoutUpload");
		QTimer::singleShot(30, myParent, SLOT(reject()) );
	}
}

void Client::explainProblem(int reason){
	QString str="";
	QString suggestedHelp=errorhelp[reason];
	if (suggestedHelp.isEmpty())
		suggestedHelp=="Don't know. Please send an idea to cye-improve@AndrasKrueger.de";
	QTextOStream(&str)<<"A problem occurred:\n"
		<<" ("<<reason<<") "<<errormessage[reason]
		<<"\nSuggested solution:\n "<<suggestedHelp;
	infoText->append(str);
	infoText->setContentsPos (0,infoText->contentsHeight());
	QMessageBox::information(this, "Some problem occurred during connection", str);
	emit Event("ClientProblemExplained");
}

void Client::TESTToggleOpenSendProcessAnswerClose(bool on){
	if (on){
		noWait=true;
		TESTsendBox->setEnabled(false);
	}
	else {
		noWait=false;
		TESTsendBox->setEnabled(true);
	}
}


void Client::setHost(){
	servername=editHostaddress->text();
	bool ok;	Q_UINT16 newPort=editHostport->text().toULong( &ok, 10);
	serverport= (ok) ? newPort:serverport;
	infoText->append(tr("Server changed to  %1:%2 ").arg(servername).arg(serverport));
	infoText->setContentsPos (0,infoText->contentsHeight());
}

void Client::openConnection(){	
	emit Event("ClientOpenConnection");
	infoText->append( tr("Trying to connect to the server...") );
	infoText->setContentsPos (0,infoText->contentsHeight());
	hostnameFound=false;
	serverConnected=false;
	DNSTimer->start(CLIENT_TIMEOUT, true);
	theSocket->connectToHost( servername, serverport );
}

void Client::socketHostFound() {
	emit Event("ClientServerDNSFound");
	DNSTimer->stop();
	// the fucking QTimer object can't be stop()ed! So I need this booleans
	hostnameFound=true; 
	infoText->append( tr("Hostname found in DNS...") );
	infoText->setContentsPos (0,infoText->contentsHeight());
	ConnectTimer->start(CLIENT_TIMEOUT, true);
}

void Client::HostNotFoundInDNS(){
	if (!hostnameFound){
		emit Event("ClientServerDNSNotFound");
		infoText->append( tr("Host not found in DNS.") );
		infoText->append( tr("Please check Internet connection and try again.") );
		infoText->setContentsPos (0,infoText->contentsHeight());
	}
}

void Client::socketConnected()   {
	emit Event("ClientServerConnected");
	ConnectTimer->stop();
	// the fucking QTimer object can't be stop()ed! So I need this boolean
	serverConnected=true;
	infoText->append( tr("Connected to server") );
	infoText->setContentsPos (0,infoText->contentsHeight());
	if (noWait) 
		sendToServer();
}

void Client::ServerNotFound(){
	emit Event("ClientServerNotFound");
	if (!serverConnected){
		infoText->append( tr("Server cannot be reached at the moment.") );
		infoText->append( tr("Please check your Internet connection and try again (later).") );
		infoText->setContentsPos (0,infoText->contentsHeight());
	}
}
    
void Client::socketConnectionClosed() {
	emit Event("ClientConnectionClosedByServer");
	infoText->append( tr("Connection closed by the server. Malfunction, please try again (later).") );
	infoText->setContentsPos (0,infoText->contentsHeight());
}
 
void Client::socketClosed(){
	emit Event("ClientConnectionClosedProperly");
	infoText->append( tr("Connection closed properly") );
	infoText->setContentsPos (0,infoText->contentsHeight());
}

void Client::socketError( int e ){
	emit Event("ClientSomeSocketError");
	serverConnected=true;
	infoText->append( tr("Socket error number %1 occurred").arg(e) );
	infoText->setContentsPos (0,infoText->contentsHeight());
	emit ClientProblem((int)TCP_NO_CONNECTION_WITH_SERVER);
}

   
void Client::closeConnection() {
	emit Event("ClientCloseConnection");
	theSocket->close();
	if ( theSocket->state() == QSocket::Closing ) { // We have a delayed close.
	    connect( theSocket, SIGNAL(delayedCloseFinished()), SLOT(socketClosed()) );
	} else {   // The socket is closed.
	    socketClosed();
	}
}

void Client::clearView(){	
	infoText->setText(""); 
	emit Event("ClientClearView");
}
    

#include "exceptions.h"

void Client::sendToServer() {	
	emit Event("ClientSendToServer");
	infoText->append( tr("starting to send to server...") );
	infoText->setContentsPos (0,infoText->contentsHeight());
	Q_UINT32 theSize;
	try{
		theSize=sizeOf(*DataStoragePointer);
		// now the streaming:
		QDataStream osdata(theSocket);
		osdata << (Q_UINT32) MAGICNUMBER;
		osdata << (Q_UINT16) PROGRAMVERSION;
		osdata << theSize;
		osdata << *DataStoragePointer ;
	}
	catch (StreamErr e) {
		closeConnection();
		infoText->append( tr("transmission not successful!") );
		infoText->setContentsPos (0,infoText->contentsHeight());
		switch(QMessageBox::warning(this, tr("CYE net error %1").arg(e.errnum), 
			errorMessageAndSolution(TCP_NET_STREAM_ERROR),
			"Try again", 
			"Quit now", 0, 0, 1 ) )
		{	case 0: 
				emit Event("ClientBadDatastreamTryAgain");
		  	  break;
			default:
				emit Event("ClientBadDatastreamQuit");
				throw NetErr(TCP_NET_STREAM_ERROR);
				// TODO: Why can't this be caught?
			  break;
		}
	}
}

#include <math.h>

void Client::TESTSendTooMuchToServer() {	
	Q_UINT32 theSize=(Q_UINT32) pow(2,31);

	QDataStream osdata(theSocket);
	osdata << (Q_UINT32) MAGICNUMBER;
	osdata << (Q_UINT16) PROGRAMVERSION;
	osdata << theSize;
}

void Client::TESTSendWrongFormatToServer()
{	
	QDataStream osdata(theSocket);
	osdata << (Q_UINT32) NO_MAGICNUMBER;
}

void Client::TESTSendOldVersionToServer(){	
	QDataStream osdata(theSocket);
	osdata << (Q_UINT32) MAGICNUMBER;
	Q_UINT16 v=10;
	osdata << v;
}

void Client::TESTSendTooNewVersionToServer(){	
	QDataStream osdata(theSocket);
	osdata << (Q_UINT32) MAGICNUMBER;
	Q_UINT16 v= 1000;
	osdata << v; 
}


void Client::socketReadyRead()  {
	infoText->append( tr("receiving from server...") );
	infoText->setContentsPos (0,infoText->contentsHeight());
	QDataStream is(theSocket);

	if (HeaderRead==false){
		Q_UINT32 magic;
		is >> magic;
		if (magic!=MAGICNUMBER){
			emit Event("ClientServerBadDataFormat");
			emit IllegalConnection((int)TCP_SERVER_HAS_BAD_DATA_FORMAT);
			return;
		}
		Q_UINT16 version;
		is >> version; // on client side, ignore mismatch
		is >> theSize;
		// qDebug(tr("size of incoming data: %1").arg(theSize) );
		HeaderRead=true;
	}
	
	Q_UINT32 bytesToWaitFor=theSocket->bytesAvailable();

	if (bytesToWaitFor == theSize){
		infoText->append( tr("all data arrived from server") );
		infoText->append( tr("---") );
		infoText->setContentsPos (0,infoText->contentsHeight());
		emit Event("ClientReceivingAnswer");
		ServerAnswer theServerAnswer;
		try {
			is >> theServerAnswer;
		}
		catch (myException e){
			emit Event("ClientServerAnswerBroken");
			emit ClientProblem(e.errnum);
			if (noWait) closeConnection();
			return;
		}

		bool ok=theServerAnswer.ProcessAnswerAndStoreInto(
			DataStoragePointer, DialogsPointer, infoText);
		if (!ok) {
			emit ClientProblem((int)TCP_SERVER_DIDNT_LIKE_CLIENT_MEMBERNUMBER);
			emit Event("ClientServerDoesntRecognizeMember");
		}
		else 
		{
			uploadSuccessful=true;
			open->setEnabled(false);
		}

 		HeaderRead=false;

		if (noWait) closeConnection();
	}
}

#include <qmessagebox.h>

void Client::CloseTheConnection(int GoodbyeReason){
	emit Event("ClientHadToCloseTheConnectionImproperly");
	// Try to read out server answer, if there is any
	QString error=tr("Client had to close the connection %1 for reason %2")
		            .arg(theSocket->socket()).arg(GoodbyeReason);
	QString line, ServerAnswerText=error+"\n\nBecause server answered:\n";
	int lines=0;
	while (theSocket->canReadLine() && lines <20){
		line="";
		QTextOStream (&line) << theSocket->readLine();
		ServerAnswerText+=line;
		lines++;
	}
	closeConnection();

	if (lines==0) {
		ServerAnswerText+="Don't know";
		QMessageBox::information( this, "Communication with Server failed", 
								  ServerAnswerText );
		infoText->append( ServerAnswerText );
		infoText->setContentsPos (0,infoText->contentsHeight());
	}
	else{
		int before = ServerAnswerText.find('(');
		int after = ServerAnswerText.find(',', before);
		bool ok;
		QString stringnumber=ServerAnswerText.mid(before+1, after-before-1);
		uint errnum=stringnumber.toUInt(&ok);
		if (ok) 
			emit ClientProblem((int)errnum);
	}
	
	/*
	// This gives a second messagebox

	if (lines==0) {
		ServerAnswerText+="Don't know";
	}
	else{
		int before = ServerAnswerText.find('(');
		int after = ServerAnswerText.find(',', before);
		bool ok;
		QString stringnumber=ServerAnswerText.mid(before+1, after-before-1);
		uint errnum=stringnumber.toUInt(&ok);
		if (ok) 
			emit ClientProblem((int)errnum);
	}
	
	QMessageBox::information( this, "Communication with Server failed", 
							  ServerAnswerText);
	infoText->append( ServerAnswerText );

    */
}

    
