// helpers.cpp
// little classes and functions which haven't found their home yet
// v0.84 (7.3.2002)
// (c) 2002 Andreas Krueger
//
// Please, if you change anything in this code,
// mark the end of those lines with '// *+* your name *+* ' 
// (so that I can find it) and send the code 
// to cye-code@AndreasKrueger.de
//
// last change: 
//          by:


#include<qtextstream.h>

#include"helpers.h"
#include"constants.h"

string encrypt(string &original){
	ulong size=original.size();
	uchar c;
	string encrypted="";
	for (ulong i=0; i<size; i++){
		c=original[i];
		if (c % 2 == 0) c=c+1;
		else c=c-1;
		c=-c + 14;
		encrypted += c;
	}
	return encrypted;
};

string decrypt(string &encrypted){
	ulong size=encrypted.size();
	uchar c;
	string decrypted="";
	for (ulong i=0; i<size; i++){
		c=encrypted[i];
		c=-( c- 14);
		if (c % 2 != 0 ) c=c-1;
		else c=c+1;
		decrypted += c;
	}
	return decrypted;
};

#include <qdatastream.h>

QDataStream& operator<< (QDataStream &s, string str){
	Q_UINT32 size =str.size();
	s << size;
	Q_UINT8 c;
	string encr=encrypt(str);
	for (Q_UINT32 i=0; i<size; i++){
		c=encr[i];
		s << c;
	}
	return s;
}
QDataStream& operator>> (QDataStream &s, string &str){
	Q_UINT32 size;
	s >> size;
	Q_UINT8 c;
	string encrd="";
	for (Q_UINT32 i=0; i<size; i++){
		s >> c;
		encrd += c;
	}
	str=decrypt(encrd);
	return s;
}

#include <qstring.h>
QTextStream& operator<< (QTextStream &s, string str){
	Q_UINT32 size =str.size();
	QChar c;
	for (Q_UINT32 i=0; i<size; i++){
		c=QChar(str[i]);
		s << c;
	}
	return s;
}

// helper for listview:

void invertOrderOf(QListViewItem *LV, QListViewItem *last){
	QListViewItem *it,*itnext;
	QListViewItem *first=LV->firstChild();

	for (it=first ; it!=last; it=itnext){
		itnext=it->nextSibling();
		it->moveItem(last);
	}
}

void invertOrder(QListViewItem *first, QListViewItem *last){
	QListViewItem *it,*itnext;
	for (it=first ; it!=last; it=itnext){
		itnext=it->nextSibling();
		it->moveItem(last);
	}
}

int minOf2(int a, int b) {return (a<b)?a:b;};
int maxOf2(int a, int b) {return (a>b)?a:b;};

cellCoord nextCellForEditing(uint r, uint c){
	cellCoord next;
	next.column=c;
	next.row=r+1;
	if (next.row>=NumCountries){
		next.row=0;
		next.column++;
	}
	if (next.column>=NumCoins)
		next.column=0;
	return next;
}

TimedCell::TimedCell(cellCoord theCell, int msec, QObject *parent=0, const char *name=0 )
  : QObject(parent, name), cell(theCell)
{
	theTimer = new QTimer(this, "cell timer");
	theTimer->start ( msec, TRUE ); // single Shot = true
	connect( theTimer, SIGNAL(timeout()), this, SLOT(timeout()) );
}

TimedCell::~TimedCell(){
	delete theTimer;
}

bool TimedCell::isActive(){
	return theTimer->isActive();
}

void TimedCell::stop(){
	theTimer->stop();
}

void TimedCell::timeout(){
	emit timeIsOut(cell);
}



AutoErasedCells::AutoErasedCells ( QObject *parent, const char *name )
		: QObject(parent, name)
{}

void AutoErasedCells::insert(cellCoord aCell){

	if (isActive(aCell))	Timers.remove(aCell);
	
	TimedCell * temp= new TimedCell(aCell, 5000, this, "timed cell");
	Timers.insert(aCell, temp);

	connect( temp, SIGNAL(timeIsOut(cellCoord)), this, SLOT(finished(cellCoord)) );

}

bool AutoErasedCells::isActive(cellCoord aCell){
		if (!Timers.contains(aCell)) return false;
		else return Timers[aCell]->isActive();
}

void AutoErasedCells::remove(cellCoord aCell){
	if (isActive(aCell))	{
		Timers[aCell]->stop();

		delete Timers[aCell];
		Timers.remove(aCell);
	}
}

void AutoErasedCells::finished(cellCoord aCell){
	delete Timers[aCell] ;
	Timers.remove(aCell);

	emit eraseCell(aCell);
}



QString errorMessageAndSolution( uint errnum ){
	QString temp="";
	temp+=errormessage[errnum];
	if (!(errorhelp[errnum]=="")){
		temp+="\nPossible solution:\n";
		temp+=errorhelp[errnum];
	}
	else{
		temp+="\n\nPlease write a short message\n"
			"with a description of this problem\n"
			"to cye-bugs@AndreasKrueger.de";
	}
	return temp;
}

#include <qfile.h>

int copyFile(QString from, QString to){
    QFile f_from( from );
    bool ok=f_from.open( IO_ReadOnly );
	if (!ok) {
		// qDebug("from-file open error in helpers::copyFile(QString from, QString to)");
		return CYE_FILE_NOT_EXIST;
	}
    
	QFile f_to( to );
    ok=f_to.open( IO_WriteOnly );
	if (!ok) {
		// qDebug("to-file open error in helpers::copyFile(QString from, QString to)");
		return CYE_FILE_WRITE_OPENING_ERROR;
	}

	int byte=f_from.getch();
	while (byte != -1){
		f_to.putch(byte);
		byte=f_from.getch();
	}

	f_from.close();
	if (f_from.status()!=IO_Ok) {
		// qDebug("Some error closing file in helpers::copyFile(QString from, QString to)");
		return CYE_FILE_CLOSING_ERROR;
	}
	f_to.close();
	if (f_to.status()!=IO_Ok) {
		// qDebug("Some error closing file in helpers::copyFile(QString from, QString to)");
		return CYE_FILE_CLOSING_ERROR;
	}
	return 0;
}

#if QT_VERSION >= 400
void openBrowser(QString target){
	QProcess *brws=new QProcess();
	brws ->addArgument(target);
	brws->start();
}
#else

#ifdef _MSC_VER 
#include <windows.h> // for ShellExecute()
void openBrowser(QString target){
	HINSTANCE h = ShellExecute(NULL, "open", 
		target, NULL,NULL, SW_SHOWNORMAL);
	if (32 >= reinterpret_cast<UINT>(h) )
				int nothing=0;// it didn't work
}
#else
#include <qmessagebox.h>

void openBrowser(QString target){
	QString text="Sorry, I don't know how to open a browser in Qt v%1 ";
	text+="that points to this page:\n       ";
	text+=target;
	text+="\n\nIf YOU know howto, please tell me at cye-code@AndreasKrueger.de";
	text=text.arg(QT_VERSION/100.0);
	QMessageBox::information(NULL, "CYE Sorry, this isn't windows", text);
}
#endif // _MSC_VER
#endif // QT version


// These two are just used to get rid of the stupid 
// g++ warning "unused variable"
int doNothing(int dummy)
{return dummy;}

QString doNothing(const QString &dummy)
{return dummy;}

// isEmail?  a@be.tv
// 7 characters, 
// one @ but not at front or back
// one . but not at back
bool isEmail(QString s){
	int length=s.length();
	if (length<7)	return false;
	int atpos=s.find('@');
	if (atpos<1 || atpos==length-1)	return false;
	int dotpos=s.find('.') ;
	if (dotpos<0 || dotpos==length-1 )	return false;
	return true;
}

bool is_it_Windows(){ // true=Windows, false=Other (_WIN32 not defined)
#ifdef _WIN32
	return true;
#else
	return false;
#endif;
}
