// storage.cpp
// data container for storing the purse contents
// v0.84 (7.3.2002)
// (c) 2002 Andreas Krueger
//
// Please, if you change anything in this code,
// mark the end of those lines with '// *+* your name *+* ' 
// (so that I can find it) and send the code 
// to cye-code@AndreasKrueger.de
//
// last change:  19.3.2002
//          by:  AK

#include"storage.h"
#include <qvaluelist.h>
#include <qtextstream.h>
#include <qapplication.h>
#include <qmessagebox.h>

// error handling:
bool TESTTHROW = false;

/////////////////////////////////////////////////////////////////////
// class CoinPile
/////////////////////////////////////////////////////////////////////


#include <math.h>
#include <float.h>

bool operator== (CoinPile &A , CoinPile &B){
	return (A.coin == B.coin && A.country==B.country && A.number==B.number);
}

#include"exceptions.h"

QValueList<Q_UINT8> CoinPile::serialize(){
	Q_UINT32 size;
	if (number==0){
		size=2; 
	}
	// one byte for the coin (front=coin and back=country), 
	// then each 7bit for the number, 1bit for overflow
	size=2+(int)(DBL_EPSILON+(double)log(number)/(double)log(128.0)); 
			
	QValueList<Q_UINT8> ByteList;

	Q_UINT8 byte=coin + (country)*NumCoins ;
	ByteList.append(byte);

	Q_UINT32 numbtemp=number;
	Q_UINT32 index=1;
	do{
		byte=numbtemp % 128;
		numbtemp=(Q_UINT32) floor(DBL_EPSILON + numbtemp/128.0);
		if (numbtemp!=0) 
			byte+=128;
		ByteList.append(byte);
		index++;
	}while (numbtemp!=0);
	if (index!=size  || index!=ByteList.count() ){
		qDebug("FATAL: something went wrong serializing in CoinPile::serialize");
		throw LogicErr(CYE_SERIALIZING_ERROR);
	}
	return ByteList;
}

CoinPile deserialize(QValueList<Q_UINT8> &ByteList){
	Q_UINT32 size=ByteList.count();	
	QValueList<Q_UINT8>::Iterator byteIter=ByteList.begin();
	int coin=*byteIter % NumCoins ;
	int country=(int) floor(*byteIter / (double)NumCoins + DBL_EPSILON ) ;
	Q_UINT32 index=1;
	Q_UINT32 number=0;
	Q_UINT32 multiplier=1;
	do{
		byteIter++;
		number+=(*byteIter % 128) * multiplier;
		multiplier*=128;
		index++;
	} while( *byteIter >=128 );
	if (index!=size  || index!=ByteList.count()) {
		qDebug("FATAL: something went wrong serializing in CoinPile::deserialize");
		throw LogicErr(CYE_DESERIALIZING_ERROR);
	}
	return CoinPile(coin, country, number);
}


void testStorageClass(){
	QValueList<Q_UINT8> tempByteList;  
	CoinPile tempPile;
	QList<ulong> randoms; randoms.setAutoDelete( TRUE );     
	ulong randnumber;

	QString str;
	for (uint coin=0; coin<NumCoins;coin++){
		for (uint country=0; country<NumCountries;country++){
			randnumber=rand();
			randoms.append(new ulong(randnumber));

			tempByteList=CoinPile(coin,country,randnumber).serialize();

			tempPile=deserialize(tempByteList);

			str="";
			QTextOStream(&str) <<"ByteList-size: "<<tempByteList.count()<<",  "<<coin<<"->"<<tempPile.coin<<"  "<<country<<"->"<<tempPile.country<<"  "<<randnumber<<"->"<<tempPile.number<<"  ";
			qDebug(str);
			if (coin!=tempPile.coin || country!=tempPile.country || randnumber!=tempPile.number) 
				qDebug("different result after serialize deserialize");
		}
	}
}

void test2StorageClass(){
	QValueList<Q_UINT8> tempByteList; 
	CoinPile tempPile;

	QString str;
	uint coin=4; uint country=7;
	long modulo=1;
	for (Q_UINT32 number=0; number<=1000000;number++){

		tempByteList=CoinPile(coin,country,number).serialize();

		tempPile=deserialize(tempByteList);

		if (number / modulo == 10) modulo*=10;

		if (number % modulo ==0) {
			str="";
			QTextOStream(&str) << number;
			qDebug(str);
			str="";
			QTextOStream(&str) <<"ByteList-size: "<<tempByteList.count()<<",  "<<coin<<"->"<<tempPile.coin<<"  "<<country<<"->"<<tempPile.country<<"  "<<number<<"->"<<tempPile.number<<"  ";
			qDebug(str);
		}

		if (coin!=tempPile.coin || country!=tempPile.country || number!=tempPile.number) {
			qDebug("different result after serialize deserialize");
			str="";
			QTextOStream(&str) <<"ByteList-size: "<<tempByteList.count()<<",  "<<coin<<"->"<<tempPile.coin<<"  "<<country<<"->"<<tempPile.country<<"  "<<number<<"->"<<tempPile.number<<"  ";
			qDebug(str);
		}
	}
}


QTextStream& operator<< (QTextStream &s, CoinPile &CP){
	return s<<"("<<CP.CoinCountryID()<<","<<CP.CoinValue()<<")="<< CP.NumberOfCoins();
}


#include <strstream>
using std::ostrstream;

char* CoinPile::giveAllCoinCountriesAsTextTupels(char* thirdcolumn){
	SettingsDialog SD;
	ostrstream ss;
	uint i;
	for (i=0; i<NumCountries+1;i++){ 
		ss<<"("<<i<<",'"<<countrynamesWords[i].latin1()<<"'"<<thirdcolumn<<"),";
	}
	ss.seekp(ss.tellp()-(long)1); // cut out the last comma
	ss<<'\0';
	//qDebug("%s",ss.str());
	return ss.str();
}

char* CoinPile::giveAllCoinValuesAsTextTripels(char* fourthcolumn){
	SettingsDialog SD;
	ostrstream ss;
	uint i;
	for (i=0; i<NumCoins;i++){ 
		ss<<"("<<i+1<<","<<coinvalues[i]<<",'"<<coinnames[i].latin1()<<"'"<<fourthcolumn<<"),";
	}
	ss.seekp(ss.tellp()-(long)1); // cut out the last comma
	ss<<'\0';
	//qDebug("%s",ss.str());
	return ss.str();
}


/////////////////////////////////////////////////////////////////////
// class Session
/////////////////////////////////////////////////////////////////////

Session::Session(){
  Version=PROGRAMVERSION;
  EntryDateTime=QDateTime::currentDateTime(); 
  SumCents=0;
  Purse.clear();
  SessionEvents.clear();
}

Session::Session(Q_UINT32 sumCents_, QDateTime datetime, 
				 Q_UINT16 progversion)
				: SumCents(sumCents_), EntryDateTime(datetime),
				  Version(progversion)
{ };

Session::Session(const Session& Source)
				: SumCents(Source.SumCents), EntryDateTime(Source.EntryDateTime),
				  Version(Source.Version)
{
	Purse=Source.Purse;
	SessionEvents=Source.SessionEvents;
}

Session& Session::operator=(Session& Source){
	qDebug("FATAL: accessed Session& Session::operator=(Session& Source)"
		   " but not implemented properly!");
	throw LogicErr(CYE_IMPLEMENTATION_CHAOS);
	return *new Session(Source);
}

void Session::insertCoins(int coin, int country, Q_UINT32 number){
	Purse.append (CoinPile(coin, country, number));
	SumCents+=Q_UINT32(number*coinvalues[coin]*100);
}
void Session::insertCoins(CoinPile cp){
	Purse.append (cp);
	SumCents+=Q_UINT32(cp.number*coinvalues[cp.coin]*100);
}

void Session::emptyPurse(){
	Purse.clear();
	SumCents=0;
}

vector<vector<Q_UINT32> > Session::PurseToCounttableArray(){
	vector<vector<Q_UINT32> > array;
	array.resize(NumCoins);
	uint coin, country;
	for (coin=0; coin<NumCoins;coin++){
		array[coin].resize(NumCountries);
		for (country=0; country<NumCountries;country++){
			array[coin][country]=0;
		}
	}
	Q_UINT32 number;
	QValueList<CoinPile>::Iterator current=Purse.begin();
	while(current!=Purse.end()){
		coin=(*current).coin;
		country=(*current).country;
		number=(*current).number;
		array[coin][country]=number;
		current++;
	}
	return array;
}

void Session::SetPurseFromCounttableArray(const vector<vector<Q_UINT32> > &array){
	emptyPurse();
	uint coin, country; Q_UINT32 number; 
	for (coin=0; coin<NumCoins;coin++){
		for (country=0; country<NumCountries;country++){
			number=array[coin][country];
			if(number!=0)
				insertCoins(coin, country, number);
		}
	}
}

QTextStream& operator<< (QTextStream &s, Session SI){
	
    s << SI.Version;
	s << SI.EntryDateTime.toString();
	s << SI.SumCents;

	Usage *theUsage =new Usage();
	theUsage->setDataTo(SI.SessionEvents);
	QString str;
	QTextOStream tos(&str);
	theUsage->PrintSessionTo(tos);
	s << str;
	delete theUsage;

	Q_UINT16 number=SI.Purse.count();
	s << number;

	QValueList<CoinPile>::Iterator current=SI.Purse.begin();
	while(current!=SI.Purse.end()){
		s << *current;
		current++;
	}
	return s;
}

QDataStream& operator<< (QDataStream &s, Session SI){

    s << (Q_UINT16) PROGRAMVERSION;
	s << SI.EntryDateTime;
	s << SI.SumCents;
	s << SI.SessionEvents;

	Q_UINT16 number=SI.Purse.count();
	s << number;

	QValueList<Q_UINT8> serialByteList; 
	QValueList<Q_UINT8>::Iterator byteIter;   Q_UINT8 byte;

	QValueList<CoinPile>::Iterator current=SI.Purse.begin();
	while(current!=SI.Purse.end()){

		serialByteList = (*current).serialize();
		
		for(byteIter=serialByteList.begin();byteIter!=serialByteList.end();byteIter++){
			byte=*byteIter;  
			s << byte;
		}
		current++;
	}
	return s;
}


QDataStream& operator>> (QDataStream &s, Session &SI){

    s >> SI.Version;

    if ( SI.Version < MINVERSION ){ 
		qDebug("version error in QDataStream& operator>> FILE TOO OLD");
		throw StreamErr(CYE_BAD_FILE_TOO_OLD);
	}
    if ( SI.Version > MAXVERSION){
		qDebug("version error in QDataStream& operator>> FILE TOO NEW");
		throw StreamErr(CYE_BAD_FILE_TOO_NEW);
	}

	s >> SI.EntryDateTime;
	Q_UINT32 checkSumCents	;
	s >> checkSumCents;

	s >> SI.SessionEvents;

	Q_UINT16 numItems;
	s >> numItems;

	SI.emptyPurse();

	QValueList<Q_UINT8> serialByteList; 
	Q_UINT8 byte;

	for (int i=0; i<numItems; i++){
		s >>byte;
		serialByteList.append(byte);
		int index=1;
		do{
			s >>byte;
			serialByteList.append(byte);
			index++;
		}while (byte>=128);
		SI.insertCoins(deserialize(serialByteList));
		serialByteList.clear();
	}
	if (checkSumCents!=SI.SumCents) {
		qDebug("Checksum differs! QDataStream& operator>> (QDataStream &s, Session &SI)");
		throw StreamErr(CYE_FILE_CORRUPT_FORMAT_SUM_DIFFERS);
	}
	return s;
}

void Session::CreateListViewItemChildrenOf(QListViewItem *LV, Usage & Observer){
	QString temp, temp2;
	QListViewItem *last=
		   new QListViewItem( LV, "Version", temp.setNum(Version) );
	(void) new QListViewItem( LV, "EntryDateTime", EntryDateTime.toString());
	(void) new QListViewItem( LV, "Sum (Euro)", temp.setNum(SumCents/100.0));

	QListViewItem *CoinPilesLV=
		   new QListViewItem( LV, "CoinPiles", temp.setNum(Purse.count()));
	if (Purse.count() != 0){
		QValueList<CoinPile>::Iterator CP=Purse.begin();
		QListViewItem *first=0, *actual;
		bool firstCoinLV=true;
		for (; CP!=Purse.end(); CP++){
			temp =coinnames[(*CP).coin];
			temp+=" "+countrynamesWords[(*CP).country+1];
			actual=new QListViewItem( CoinPilesLV, temp, temp2.setNum((*CP).number));
			if (firstCoinLV){
				firstCoinLV=false;
				first=actual;
			}
		}
		invertOrderOf(CoinPilesLV, first);
	}

	QListViewItem * sessionEventsLV =
		new QListViewItem( LV, "SessionObservation");
	Observer.setDataTo(SessionEvents);

	Q_UINT32 TotalNumEvents=
		Observer.SessionEventsCreateListViewItemChildrenOf(sessionEventsLV);
	sessionEventsLV ->setText(1, temp.setNum(TotalNumEvents));

	invertOrderOf(LV, last);
}


bool operator== (Session &A, Session &B){
	// first check, if the CoinPiles are identical (and in the same order)
	QValueList<CoinPile>::Iterator a, b;  
	a= A.Purse.begin(); b= B.Purse.begin();
	bool temp=true;
	while( a != A.Purse.end() && b != B.Purse.end()){
		if (! (*a == *b)) temp = false;
		a++; b++;
	}
	if (a!=A.Purse.end() || b!=B.Purse.end()) temp=false;

	// Then check, if the SessionEvents are identical
	QValueList<UserEvent>::Iterator a2, b2;  
	a2= A.SessionEvents.begin(); b2= B.SessionEvents.begin();

	while( a2 != A.SessionEvents.end() && b2 != B.SessionEvents.end()){
		if (! (*a2 == *b2)) temp = false;
		a2++; b2++;
	}
	if (a2!=A.SessionEvents.end() || b2!=B.SessionEvents.end()) 
		temp=false;
	
	// now check all the other variables:
	temp = temp && A.EntryDateTime == B.EntryDateTime ;
	temp = temp && A.Version==B.Version;
	return temp;
}


vector<vector<Q_UINT32> > test3StorageClass(const vector<vector<Q_UINT32> > &CountTableArray){

	Session testItem(0, QDateTime(QDate(2002,1,31), QTime(1,2,3,4)), MINVERSION);
	Session compareItem(1, QDateTime::currentDateTime(), MINVERSION-1);
	testItem.SetPurseFromCounttableArray(CountTableArray);
	testItem.SessionEvents.append(UserEvent(12, 13));
	testItem.SessionEvents.append(UserEvent(15, 14));

	QFile f( FILENAME );
    f.open( IO_WriteOnly );
    QDataStream s( &f );
	s << testItem;
	f.close();
	if(f.status()!=IO_Ok) 
		qDebug("error saving");

	f.open( IO_ReadOnly );
	s >> compareItem;
	f.close();
	if(f.status()!=IO_Ok) 
		qDebug("error loading");

	if (! (testItem == compareItem)){
		QString str;
		QTextOStream (&str) << "different before and after saving & loading\n";
		qDebug(str);
		throw LogicErr(CYE_TESTING_ERROR);
	}

	vector<vector<Q_UINT32> > cta=	compareItem.PurseToCounttableArray();
	return cta;
}


/////////////////////////////////////////////////////////////////////
// class ClientStorage
/////////////////////////////////////////////////////////////////////

ClientStorage::ClientStorage() {
	// do that manually after creating(!): setDefaultValues();
}

void ClientStorage::setDefaultValues(){
	Version=PROGRAMVERSION; 
	Membernumber=0;
	Welcome="";
	RecommenderMembernumbers.clear();
	RecommenderName="";
	Country=0;
	PostalCode="";
	CityPopulation=0;
	eMailAddress="";
	eMailFlags=0;
	SettingsFlags=256; // Settings have to be stored (in the server database)
	// store the operating system:
	SettingsFlags |= (is_it_Windows()?0:8);

	MemberSince=QDateTime (QDate(2002,2,1),QTime(0,0,0));
	NumberOfUploads=0;
	LastUpload=MemberSince;
	NumberOfStartsSinceLastUpload=0;

	srand(QTime::currentTime().msecsTo(QTime(0,0,0)));
	// choose a random favourite coin for the first start:
	FavouriteCoin=(Q_UINT8)(((double)rand()/RAND_MAX)*(NumCountries*NumCoins));
	GuessWhenMixdateYear=0; GuessWhenMixdateMonth=0;
	GuessWhenArrivaldateYear=0; GuessWhenArrivaldateMonth=0;
	Message="";

	ServerAddress=SERVERNAME;
 	ServerPort=(Q_UINT16) SERVERPORT;
	FTPServer=FTPSERVERNAME;

	NumberOfPicturesAvailable=0;
	LastAnalysisPage=0;
	LastTxtMessage=0;
	LastPicture=0;
	LastDataAnalysisRead=MemberSince;

	WPosition=QPoint(0,0);
    WSize=QSize(600,400);

	TotalEvents.clear();
	NumFullPurses=0;
	Sessions.clear();

	InvitationList.clear();
}

ClientStorage::ClientStorage(const ClientStorage & S){
	this->NumFullPurses=0;
	this->Sessions.clear();

	this->MergeWithNewSessions(S); // copies all members
	                               // and appends Sessions and Purses
}

void ClientStorage::MergeWithNewSessions(const ClientStorage & S){
	// overwrites with new data from S,
	// but the sessions are merged (appended) 
	Version=S.Version;
    Membernumber=S.Membernumber;
	Welcome=S.Welcome;
    RecommenderMembernumbers=S.RecommenderMembernumbers;
	RecommenderName=S.RecommenderName;
    Country=S.Country;
    PostalCode=S.PostalCode;
	CityPopulation=S.CityPopulation;
    eMailAddress=S.eMailAddress;
	eMailFlags=S.eMailFlags;
	SettingsFlags=S.SettingsFlags;
    MemberSince=S.MemberSince;
    NumberOfUploads=S.NumberOfUploads;
    LastUpload=S.LastUpload;
    NumberOfStartsSinceLastUpload=S.NumberOfStartsSinceLastUpload;
	FavouriteCoin=S.FavouriteCoin;
	GuessWhenMixdateYear=S.GuessWhenMixdateYear;
	GuessWhenMixdateMonth=S.GuessWhenMixdateMonth;
	GuessWhenArrivaldateYear=S.GuessWhenArrivaldateYear;
	GuessWhenArrivaldateMonth=S.GuessWhenArrivaldateMonth;
	Message=S.Message;

    ServerAddress=S.ServerAddress;
    ServerPort=S.ServerPort;
	FTPServer=S.FTPServer;
    NumberOfPicturesAvailable=S.NumberOfPicturesAvailable;
    LastAnalysisPage=S.LastAnalysisPage;
    LastTxtMessage=S.LastTxtMessage;
    LastDataAnalysisRead=S.LastDataAnalysisRead;
	LastPicture=S.LastPicture;
    WPosition=S.WPosition;
    WSize=S.WSize;

	TotalEvents=S.TotalEvents;

	NumFullPurses	+= S.NumFullPurses;
	Sessions		+= S.Sessions;

	InvitationList   = S.InvitationList;
}

void ClientStorage::CreateListViewItemChildrenOf(QListViewItem *LV, Usage & Observer){
	QString temp;

	QListViewItem *last=
		   new QListViewItem( LV, "Membernumber", temp.setNum(Membernumber) );

	(void) new QListViewItem( LV, "NumFullPurses", temp.setNum(NumFullPurses));
	QListViewItem *sessLV, *oneSessLV;

	// The Sessions
	sessLV=new QListViewItem( LV, "Sessions", temp.setNum(Sessions.count()));
	QValueList<Session>::Iterator sess=Sessions.begin();
	int counter=0;
	for (; sess!=Sessions.end(); sess++){
		counter++;
		temp ="%1 euro / %2 events";
		temp = temp.arg((*sess).SumCents/100.0).arg((*sess).SessionEvents.count());
		oneSessLV=new QListViewItem( sessLV, (*sess).EntryDateTime.toString(), temp);
		(*sess).CreateListViewItemChildrenOf(oneSessLV, Observer);
	}

	// the TotalEvents
	QListViewItem *TotalEventsLV=
		new QListViewItem( LV, "TotalEvents");
	Observer.setDataTo(TotalEvents);
	Q_UINT32 TotalNumEvents=
		Observer.AllEventsCreateListViewItemChildrenOf(TotalEventsLV);
	TotalEventsLV->setText(1, temp.setNum(TotalNumEvents));

	(void) new QListViewItem( LV, "Version", temp.setNum(Version) );
	(void) new QListViewItem( LV, "Welcome", QString(Welcome.c_str()) );
	QListViewItem *rmn=
		new QListViewItem( LV, "RecommenderMembernumbers", temp.setNum(RecommenderMembernumbers.count()));
	for (QValueList<Q_UINT32>::Iterator it= RecommenderMembernumbers.begin();
	it!=RecommenderMembernumbers.end(); it++){
		(void) new QListViewItem( rmn, temp.setNum(*it));
	}
	(void) new QListViewItem( LV, "RecommenderName", RecommenderName);
	(void) new QListViewItem( LV, "Country", temp.setNum(Country));
	(void) new QListViewItem( LV, "PostalCode", PostalCode);
	(void) new QListViewItem( LV, "CityPopulation", temp.setNum(CityPopulation));
	(void) new QListViewItem( LV, "eMailAddress", QString(eMailAddress.c_str()));
	
	QListViewItem *es =
		new QListViewItem( LV, "eMailFlags", temp.setNum(eMailFlags));
	if (eMailFlags & 1) (void)new QListViewItem(es, "MonthlyReminder", temp.setNum(1));
	if (eMailFlags & 2) (void)new QListViewItem(es, "BiMonthlyReminder", temp.setNum(2));
	if (eMailFlags & 4) (void)new QListViewItem(es, "WeeklyReminder", temp.setNum(4));
	if (eMailFlags & 8) (void)new QListViewItem(es, "WantsReminder", temp.setNum(8));
	else (void)new QListViewItem(es, "WantsNo!Reminder");
	if (eMailFlags & 16) (void)new QListViewItem(es, "WantsNewsletter", temp.setNum(16));
	if (eMailFlags & 32) (void)new QListViewItem(es, "WantsToHelp", temp.setNum(32));
	if (eMailFlags & 64) (void)new QListViewItem(es, "WantsToAnalyze", temp.setNum(64));

	QListViewItem *sf =
		new QListViewItem( LV, "SettingsFlags", temp.setNum(SettingsFlags));
	if (SettingsFlags & 1) (void)new QListViewItem(sf, "ServiceTabOn", temp.setNum(1));
	if (SettingsFlags & 2) (void)new QListViewItem(sf, "OpinionTabOn", temp.setNum(2));
	if (SettingsFlags & 4) (void)new QListViewItem(sf, "NecessFilledIn", temp.setNum(4));
	if (SettingsFlags & 8) (void)new QListViewItem(sf, "DontNagConnection", temp.setNum(8));
	if (SettingsFlags & 16) (void)new QListViewItem(sf, "HasSeenSettings", temp.setNum(16));
	if (SettingsFlags & 32) (void)new QListViewItem(sf, "NoLoadPicsNagger", temp.setNum(32));
	if (SettingsFlags & 64) (void)new QListViewItem(sf, "SignedContract", temp.setNum(64));
	if (SettingsFlags & 128) (void)new QListViewItem(sf, "AllowsObservation", temp.setNum(128));
	if (SettingsFlags & 256) (void)new QListViewItem(sf, "SettingsChanged", temp.setNum(256));

	(void) new QListViewItem( LV, "MemberSince", MemberSince.toString());
	(void) new QListViewItem( LV, "NumberOfUploads", temp.setNum(NumberOfUploads));
	(void) new QListViewItem( LV, "LastUpload", LastUpload.toString());
	(void) new QListViewItem( LV, "NumberOfStartsSinceLastUpload", temp.setNum(NumberOfStartsSinceLastUpload));
	(void) new QListViewItem( LV, "FavouriteCoin", temp.setNum(FavouriteCoin));
	(void) new QListViewItem( LV, "GuessWhenMixdateYear", temp.setNum(GuessWhenMixdateYear));
	(void) new QListViewItem( LV, "GuessWhenMixdateMonth", temp.setNum(GuessWhenMixdateMonth));
	(void) new QListViewItem( LV, "GuessWhenArrivaldateYear", temp.setNum(GuessWhenArrivaldateYear));
	(void) new QListViewItem( LV, "GuessWhenArrivaldateMonth", temp.setNum(GuessWhenArrivaldateMonth));
	(void) new QListViewItem( LV, "Message", Message);
	(void) new QListViewItem( LV, "ServerAddress", QString(ServerAddress.c_str()));
	(void) new QListViewItem( LV, "ServerPort", temp.setNum(ServerPort));
	(void) new QListViewItem( LV, "FTPServer", QString(FTPServer.c_str()));
	(void) new QListViewItem( LV, "NumberOfPicturesAvailable", temp.setNum(NumberOfPicturesAvailable));
	(void) new QListViewItem( LV, "LastAnalysisPage", temp.setNum(LastAnalysisPage));
	(void) new QListViewItem( LV, "LastTxtMessage", temp.setNum(LastTxtMessage));
	(void) new QListViewItem( LV, "LastPicture", temp.setNum(LastPicture));
	(void) new QListViewItem( LV, "LastDataAnalysisRead", LastDataAnalysisRead.toString());
	temp="";
	QTextOStream(&temp) <<"("<<WPosition.x()<<","<<WPosition.y()<<","<<WSize.width()<<","<<WSize.height()<<")";
	(void) new QListViewItem( LV, "WindowPositionSize", temp);
	(void) new QListViewItem( LV, "InvitationList", "TODO: Listview not programmed yet.");
	
	invertOrderOf(LV, last);
}

bool ClientStorage::NecessarySettingsOK(){ return (SettingsFlags & 4); }
void ClientStorage::setNecessarySettingsOK(bool ok){
	SettingsFlags = (ok) ? (SettingsFlags | 4):(SettingsFlags & (65535-4));
}


void toStream  (QDataStream &s, QValueList<Session> &L){
	s << (Q_UINT16) L.count();
	QValueList<Session>::Iterator SI;
	for (SI=L.begin(); SI!=L.end(); SI++){
		s << *SI;
	}
}
void toStream  (QTextStream &s, QValueList<Session> &L){
	s << (Q_UINT16) L.count();
	QValueList<Session>::Iterator SI;
	for (SI=L.begin(); SI!=L.end(); SI++){
		s << *SI;
	}
}
void InitializeFromStream(QDataStream &s, QValueList<Session> &L){
	L.clear();
	Q_UINT16 NumberOfItems;
	s >> NumberOfItems;

	Session temp;
	for (Q_UINT16 i=0; i<NumberOfItems; i++){
		s >> temp;
		L.append(Session(temp)); 
	}
}

QDataStream& operator<< (QDataStream &s, ClientStorage S){

	s << (Q_UINT16) PROGRAMVERSION;
	s << S.Membernumber;
	s << S.Welcome;
	s << S.RecommenderMembernumbers << S.RecommenderName;
	s << S.Country <<S.PostalCode;
	s << S.CityPopulation;
	s << S.eMailAddress << S.eMailFlags;
	s << S.SettingsFlags;

	s << S.MemberSince ;
	s << S.NumberOfUploads << S.LastUpload;
	s << S.NumberOfStartsSinceLastUpload;
	s << S.FavouriteCoin;
	s << S.GuessWhenMixdateYear << S.GuessWhenMixdateMonth;
	s << S.GuessWhenArrivaldateYear << S.GuessWhenArrivaldateMonth;
	s << S.Message;

	s << S.ServerAddress << S.ServerPort;
	s << S.FTPServer; 

	s << S.NumberOfPicturesAvailable;
	s << S.LastAnalysisPage;
	s << S.LastTxtMessage << S.LastPicture;
	s << S.LastDataAnalysisRead;

	s << S.WPosition << S.WSize;

	s << S.TotalEvents;

	s << S.NumFullPurses;
	toStream(s, S.Sessions);

	s << S.InvitationList;
	return s;
}

QTextStream& operator<< (QTextStream &s, ClientStorage S){

	s << S.Version << S.Membernumber;
	s << S.Welcome;
	for (QValueList<Q_UINT32>::Iterator it= S.RecommenderMembernumbers.begin();
	                                    it!=S.RecommenderMembernumbers.end(); it++){
		s << *it;
	}

	s << S.RecommenderName;
	s << S.Country <<S.PostalCode;
	s << S.CityPopulation;
	s << S.eMailAddress << S.eMailFlags;
	s << S.SettingsFlags;

	s << S.MemberSince.toString() ;
	s << S.NumberOfUploads << S.LastUpload.toString();
	s << S.NumberOfStartsSinceLastUpload;
	s << S.FavouriteCoin;
	s << S.GuessWhenMixdateYear << S.GuessWhenMixdateMonth;
	s << S.GuessWhenArrivaldateYear << S.GuessWhenArrivaldateMonth;
	s << S.Message;

	s << S.ServerAddress << S.ServerPort;
	s << S.FTPServer; 

	s << S.NumberOfPicturesAvailable;
	s << S.LastAnalysisPage;
	s << S.LastTxtMessage << S.LastPicture;
	s << S.LastDataAnalysisRead.toString();

	s << S.WPosition.x() << S.WPosition.y() ;
	s << S.WSize.width() << S.WSize.height();

	qDebug("ATTENTION: QTextStream& operator<< (QTextStream &s, ClientStorage S)" 
		   "\n   should ONLY be used for debugging "
		   "(creates expensive temporary Usage object) !");
	Usage *theUsage =new Usage();
	theUsage->setDataTo(S.TotalEvents);
	QString str;
	QTextOStream tos(&str);
	theUsage->PrintTotalEventsTo(tos);
	s << str;
	delete theUsage;

	s << S.NumFullPurses;
	toStream(s, S.Sessions);

	s << S.InvitationList;
	
	return s;
}


QDataStream& operator>> (QDataStream &s, ClientStorage &S){
	// Principle: Whenever there is VERSION CONVERSION,
	// it happens during READING the file/datastream

	s >> S.Version ;

    if ( S.Version < MINVERSION ){ 
		qDebug("version error in QDataStream& operator>> (ClientStorage &S) FILE TOO OLD");
		throw StreamErr(CYE_BAD_FILE_TOO_OLD);
	}
    if ( S.Version > MAXVERSION){
		qDebug("version error in QDataStream& operator>> (ClientStorage &S) FILE TOO NEW");
		throw StreamErr(CYE_BAD_FILE_TOO_NEW);
	}

	s >> S.Membernumber ;
	s >> S.Welcome;
	s >> S.RecommenderMembernumbers ;
	s >> S.RecommenderName;
	s >> S.Country >>S.PostalCode >> S.CityPopulation;
	s >> S.eMailAddress >> S.eMailFlags;
	s >> S.SettingsFlags;

	// correct the v105 error: 
	// Greece was missing in SettingsDialog-> Country
	if (S.Version <=105 && ( (S.SettingsFlags & 512) == 0 ) ){
		S.SettingsFlags |= 512;  // set flag to have done it ONCE !
		if (S.Country >= 6) S.Country ++;  // to make space for Greece
		                                   // between Germany(5) and Ireland(6 -> 7)
		S.SettingsFlags |= 256;  // set flag to have server update tbl_member_private!
		                         // and insert a new MemberLocationID into tbl_member
		                         //      if country has changed here
	}

	s >> S.MemberSince ;
	s >> S.NumberOfUploads >> S.LastUpload;
	s >> S.NumberOfStartsSinceLastUpload;
	s >> S.FavouriteCoin;
	s >> S.GuessWhenMixdateYear >> S.GuessWhenMixdateMonth;
	s >> S.GuessWhenArrivaldateYear >> S.GuessWhenArrivaldateMonth;
	s >> S.Message;

	s >> S.ServerAddress >> S.ServerPort >> S.FTPServer; 

	s >> S.NumberOfPicturesAvailable;
	s >> S.LastAnalysisPage;
	s >> S.LastTxtMessage >> S.LastPicture;
	s >> S.LastDataAnalysisRead;

	s >> S.WPosition >> S.WSize;

	s >> S.TotalEvents;

	s >> S.NumFullPurses;
	InitializeFromStream(s, S.Sessions);

	s >> S.InvitationList;

	return s;
}


int ClientStorage::saveToDisk(QString filename){
    QFile f( filename );
    f.open( IO_WriteOnly );
	if (f.status()!=IO_Ok) {
		qDebug("Some error opening file for writing in ClientStorage::saveToDisk");
		return CYE_FILE_WRITE_OPENING_ERROR;
	}

    QDataStream s( &f );

	try{
		// Write a header with a "magic number" and a version
		s << (Q_UINT32) MAGICNUMBER;
		s << (Q_UINT16) PROGRAMVERSION;

		// Write the data
		s << *this;
	}
	catch (StreamErr e){ 
		f.remove();  // remove the file to prevent problems after restart
		throw FileErr(e.errnum);
	}
	catch (...){ 
		f.remove();  // remove the file to prevent problems after restart
		throw;
	}

	f.close();
	if (f.status()!=IO_Ok) {
		qDebug("Some error closing file in ClientStorage::saveToDisk");
		return CYE_FILE_CLOSING_ERROR;
	}
	return 0;
}

int ClientStorage::loadFromDisk(QString filename){
    QFile f( filename );
    bool ok=f.open( IO_ReadOnly );
	if (!ok) {
		// qDebug("file open error in ClientStorage::loadFromDisk");
		return CYE_FILE_NOT_EXIST;
	}
    QDataStream s( &f );

    // Read and check the header
    Q_UINT32 magic;
    s >> magic;
    if ( magic != MAGICNUMBER )
        return CYE_BAD_FILE_FORMAT;

    // Read the version
	Q_UINT16 VersionOnDisk;
    s >> VersionOnDisk;
	// TODO: in later versions:
	//       react to old datafiles properly!
    if ( VersionOnDisk < MINVERSION )
        return CYE_BAD_FILE_TOO_OLD;
    if ( VersionOnDisk > MAXVERSION )
        return CYE_BAD_FILE_TOO_NEW;
    
	// Read the data
	try{
		s >> *this ;
	}
	catch (StreamErr e){ 
		throw FileErr(e.errnum);
	}
	catch (...){ 
		throw;
	}

	// if (this->Version != PROGRAMVERSION){
		// 	this->Version =(Q_UINT16) PROGRAMVERSION;
	//}

	f.close();
	if (f.status()!=IO_Ok) {
		qDebug("Some error closing file in ClientStorage::loadFromDisk");
		return CYE_FILE_CLOSING_ERROR;
	}
	return 0;
}

void test4StorageClassSAVE(vector<vector<Q_UINT32> > array){
	ClientStorage *theData= new ClientStorage();
	theData->setDefaultValues();
	Session onePurse;

	onePurse.SetPurseFromCounttableArray(array);
	theData->Sessions.append(onePurse);
	theData->saveToDisk(FILENAME);
}


vector<vector<Q_UINT32> > test4StorageClassLOAD(){
		// memory leak!
		// many times loading from disk results
		// in 12 kB more memory per read-act
	ClientStorage *theData= new ClientStorage();
	theData->loadFromDisk(FILENAME);
	Session onePurse=theData->Sessions.last();
	vector<vector<Q_UINT32> > cta=onePurse.PurseToCounttableArray();
	delete theData;
	return cta;
}


void test5StorageClass(){
	// test for insertion of CoinPiles with zero number (empty piles);
	// this saves a file with half of the piles empty, 
	// just for checking, if there are problems reading that in

	Session testItem(1, QDateTime(QDate(2002,1,31), QTime(1,2,3,4)), 26);

	QValueList<Q_UINT8> tempByteList; //tempByteList.setAutoDelete( TRUE );
	CoinPile tempPile;

	QString str;
	uint coin; uint country;
	//long modulo=1;
	ulong number;
	for (int counter=0; counter<=50;counter++){
		coin=(uint) (int) (rand()/(double)RAND_MAX * NumCoins);
		country=(uint) (int) (rand()/(double)RAND_MAX * NumCountries);
		number=(ulong) (rand()>RAND_MAX/2)?1:0; // should be 0 every second try

		tempByteList=CoinPile(coin,country,number).serialize();
		tempPile=deserialize(tempByteList);

		if (coin!=tempPile.coin || country!=tempPile.country || number!=tempPile.number) {
			qDebug("different result after serialize deserialize");
			str="";
			QTextOStream(&str) <<"ByteList-size: "<<tempByteList.count()<<",  "<<coin<<"->"<<tempPile.coin<<"  "<<country<<"->"<<tempPile.country<<"  "<<number<<"->"<<tempPile.number<<"  ";
			qDebug(str);
		}
		testItem.Purse.append(CoinPile(coin, country, number));
	}
	ClientStorage fileStorage;
	fileStorage.setDefaultValues();
	fileStorage.Sessions.append(testItem);

	fileStorage.saveToDisk(FILENAME);
}

Q_UINT32 sizeOf(ClientStorage & Data){      

	// check the size by writing to dummy stream
	QByteArray  dummy;
	QDataStream dummyStream( dummy, IO_WriteOnly ); 
	dummyStream << Data;
	Q_UINT32 theSize=dummy.size();
	// how to destroy the array properly? Is that enough?
	dummy.resize(0);
	return theSize;
}

Q_UINT32 ClientStorage::totalSumOfCents(){
	QValueList<Session>::Iterator S;
	Q_UINT32 sum = 0;
	for (S=Sessions.begin(); S!=Sessions.end(); S++){
		sum += (*S).SumCents;
	}
	return sum;
}


/////////////////////////////////////////////////////////////////////
// class ServerAnswer
/////////////////////////////////////////////////////////////////////

ServerAnswer::ServerAnswer (){};

ServerAnswer::ServerAnswer (const ServerAnswer& SA){
	Version=SA.Version;
	AnswerFlags=SA.AnswerFlags;
	ServerAddress=SA.ServerAddress;
	ServerPort=SA.ServerPort;
	FTPServer=SA.FTPServer;
	DownloadServer=SA.DownloadServer;
	NewMemberNumber=SA.NewMemberNumber;
	NewMemberHello=SA.NewMemberHello;
	MemberSince=SA.MemberSince;
	NewVersion=SA.NewVersion;
	NumberOfNewestText=SA.NumberOfNewestText;
	NumberOfNewestPic=SA.NumberOfNewestPic;
	NumberOfNewestPage=SA.NumberOfNewestPage;
	NextSessionRecommendation=SA.NextSessionRecommendation;
	NumberOfTotalMembers=SA.NumberOfTotalMembers;
	NumberOfTotalPursesOnServer=SA.NumberOfTotalPursesOnServer;
	TotalSumCents=SA.TotalSumCents;
	LastUploadWasFromCountry=SA.LastUploadWasFromCountry;
	DailyMessage=SA.DailyMessage;
	MemberSpecificMessage=SA.MemberSpecificMessage;
}

QDataStream & operator<< (QDataStream& s, ServerAnswer SA){
	s << (Q_UINT16) PROGRAMVERSION;
	s << SA.AnswerFlags;
	s << SA.ServerAddress << SA.ServerPort << SA.FTPServer;
	s << SA.DownloadServer;
	s << SA.NewMemberNumber << SA.NewMemberHello;
	s << SA.MemberSince;
	s << SA.NewVersion;
	s << SA.NumberOfNewestText << SA.NumberOfNewestPic;
	s << SA.NumberOfNewestPage;
	s << SA.NextSessionRecommendation;
	s << SA.NumberOfTotalMembers << SA.NumberOfTotalPursesOnServer;
	s << SA.TotalSumCents;
	s << SA.LastUploadWasFromCountry;
	s << SA.DailyMessage;
	s << SA.MemberSpecificMessage;
	return s;
}
QTextStream & operator<< (QTextStream& s, ServerAnswer SA){
	qDebug("Why is this called? Shouldn't be used normally: QTextStream & operator<< (QTextStream& s, ServerAnswer SA)");
	s << SA.Version;
	s << SA.AnswerFlags;
	s << SA.ServerAddress << SA.ServerPort << SA.FTPServer;
	s << SA.DownloadServer;
	s << SA.NewMemberNumber << SA.NewMemberHello;
	s << SA.MemberSince.toString();
	s << SA.NewVersion;
	s << SA.NumberOfNewestText << SA.NumberOfNewestPic;
	s << SA.NumberOfNewestPage;
	s << SA.NextSessionRecommendation.toString();
	s << SA.NumberOfTotalMembers << SA.NumberOfTotalPursesOnServer;
	s << SA.TotalSumCents;
	s << SA.LastUploadWasFromCountry;
	s << SA.DailyMessage;
	s << SA.MemberSpecificMessage;
	return s;
}

QDataStream & operator>> (QDataStream& s, ServerAnswer & SA){
	s >> SA.Version;

	if ( SA.Version < MINVERSION ){ 
		qDebug("version error in QDataStream& operator>> (ServerAnswer &SA) ANSWER TOO OLD FORMAT");
		throw StreamErr(TCP_SERVER_ANSWER_TOO_OLD_FORMAT);
	}
    if ( TESTTHROW || SA.Version > MAXVERSION){
		qDebug("version error in QDataStream& operator>> (ServerAnswer &SA) ANSWER TOO NEW FORMAT");
		throw StreamErr(TCP_SERVER_ANSWER_TOO_NEW_FORMAT);
	}

	s >> SA.AnswerFlags;
	s >> SA.ServerAddress >> SA.ServerPort >> SA.FTPServer;
	s >> SA.DownloadServer;
	s >> SA.NewMemberNumber >> SA.NewMemberHello;
	s >> SA.MemberSince;
	s >> SA.NewVersion;
	s >> SA.NumberOfNewestText >> SA.NumberOfNewestPic;
	s >> SA.NumberOfNewestPage;
	s >> SA.NextSessionRecommendation;
	s >> SA.NumberOfTotalMembers >> SA.NumberOfTotalPursesOnServer;
	s >> SA.TotalSumCents;
	s >> SA.LastUploadWasFromCountry;
	s >> SA.DailyMessage;
	s >> SA.MemberSpecificMessage;
	return s;
}


#include <qtextview.h>
#include <qcombobox.h>

bool ServerAnswer::ProcessAnswerAndStoreInto(ClientStorage * theData, 
											 Dialogs * theDialogs,
											 QTextView *infoText){
	QString temp;
	if ((AnswerFlags & 1) == 0){
		infoText->append("Some error in Processing Answer - OK flag was not set.");
		infoText->append("This can have many reasons. Perhaps, your member number is wrong");
		infoText->append("- or there's some server maintenance going on!");
		infoText->append("I suggest that you try to upload again later or tomorrow, ");
		infoText->append("if the problem persists, please contact the author");
		temp="or backup and delete your datafile '%1'";
		temp=temp.arg(FILENAME);
		infoText->append(temp);
		infoText->append("---");
		infoText->setContentsPos (0,infoText->contentsHeight());
		return false;
	}
	else{
		if (AnswerFlags & 16){
			theData->Membernumber=NewMemberNumber;
			theData->Welcome=NewMemberHello;
			theData->MemberSince=MemberSince;
			infoText->append(QString(NewMemberHello.c_str()));
			infoText->append("---");
			//theDialogs->ShowNewMemberWindow(NewMemberHello, NewMemberNumber);
		}
		else {
			if (MemberSpecificMessage==""){
				temp="The Server says: 'Thanks, member %1'";
				temp=temp.arg(theData->Membernumber);
				infoText->append(temp);
			}
			else{
				temp="A personal message for you, member %1:";
				temp=temp.arg(theData->Membernumber);
				infoText->append(temp);
				infoText->append(MemberSpecificMessage.c_str());
			}
			infoText->append("---");
		}
		
		if (AnswerFlags & 2){
			theData->ServerAddress=ServerAddress;
			theData->ServerPort=ServerPort;
			theData->FTPServer=FTPServer;
			infoText->append("Minor update successful.");
		} 
		if (AnswerFlags & 12 ){
			temp="Your programversion is %1, there is a new version %2 out now.";
			temp=temp.arg(theData->Version).arg(NewVersion);
			infoText->append(temp);
			if (AnswerFlags & 4 )
				infoText->append("You should consider to update soon.");
			if (AnswerFlags & 8 )
				infoText->append("You have to update, this program version is not appropriate anymore.");
			temp="\nPlease go to %3";temp=temp.arg(DownloadServer.c_str());
			infoText->append(temp);
			infoText->append("---");
		}
		
		if (AnswerFlags & 32){
			infoText->append("New Analysis Data available on the Server: ");
			temp=" %1 pages, %2 text messages and %3 pictures\n";
			temp=temp.arg(NumberOfNewestPage - theData->LastAnalysisPage)
				     .arg(NumberOfNewestText - theData->LastTxtMessage)
				     .arg(NumberOfNewestPic - theData->LastPicture);
			infoText->append(temp);

			theData->LastTxtMessage=NumberOfNewestText;
			theData->LastPicture=NumberOfNewestPic;
			theData->LastAnalysisPage=NumberOfNewestPage;
		}
		infoText->append("---");
		infoText->append("The actual status on the server:");
		temp="# members:  %1"; temp=temp.arg(NumberOfTotalMembers);
		infoText->append(temp);
		temp="# purses :  %1"; temp=temp.arg(NumberOfTotalPursesOnServer);
		infoText->append(temp);
		temp="Total Uploaded Sum:  %1 Euro"; temp=temp.arg(TotalSumCents/100.0);
		infoText->append(temp);
		temp="last upload from: %1";
		temp=temp.arg(theDialogs->Settings->Country->text(LastUploadWasFromCountry));
		infoText->append(temp);
		temp ="suggested next session around: %1";
		temp=temp.arg(NextSessionRecommendation.toString());
		infoText->append(temp);
		infoText->append("---");
		if (DailyMessage!=""){
			infoText->append(DailyMessage.c_str());
			infoText->append("---");
		}
		infoText->setContentsPos (0,infoText->contentsHeight());
		return true;
	}
}

Q_UINT32 sizeOf(ServerAnswer & SA){
// check the size by writing to dummy stream
	QByteArray  dummy;
	QDataStream dummyStream( dummy, IO_WriteOnly ); 
	dummyStream << SA;
	Q_UINT32 theSize=dummy.size();
	// how to destroy the array properly? Is that enough?
	dummy.resize(0);
	return theSize;
}


