// storage.h
// data container for storing the purse contents
// v0.84 (7.3.2002)
// (c) 2002 Andreas Krueger
//
// Please, if you change anything in this code,
// mark the end of those lines with '// *+* your name *+* ' 
// (so that I can find it) and send the code 
// to cye-code@AndreasKrueger.de
//
// last change: 19.3.2002
//          by: AK

#ifndef STORAGE_H
#define STORAGE_H

#include "helpers.h"

#include<qwidget.h>
#include<qarray.h>
#include<qdatetime.h>
#include<qdatastream.h>
#include <qfile.h> 
#include <qmap.h> 

#include"constants.h"

/////////////////////////////////////////////////////////////////////
// class CoinPile
//
// can hold the number of one type of coin
// the number is stored in 7bit+ overflow
// way, so that any number can be stored, 
// just the number of bytes rises
// usually (<128): 1 byte cointype, 1 byte number
// 
/////////////////////////////////////////////////////////////////////

#include<qvaluelist.h>
#include<qtextstream.h>

class CoinPile {
public:
	uint coin;		// 0 - 7						
	uint country;	// 0 - NumCountries (14)
	Q_UINT32 number;
	Q_UINT8 CoinCountryID()   {return 1+country;}; // 1-15 1=Austria, 15=Vatican
	Q_UINT8 CoinValueID()     {return 1+coin;};	   // 1-8	1=1cent,    8=2euro
	double CoinValue() {return coinvalues[coin];}; // 0.01 - 2.00
	Q_UINT32 NumberOfCoins(){return number;};


public:
	CoinPile(int theCoin=0, int theCountry=0, Q_UINT32 theNumber=0)
		: coin(theCoin), country(theCountry), number(theNumber){};

	QValueList<Q_UINT8> serialize();
	friend CoinPile deserialize(QValueList<Q_UINT8> &ByteList);
	friend QTextStream& operator<< (QTextStream &s, CoinPile &CP);

	// testing:
	friend void testStorageClass();
	friend void test2StorageClass();
	friend bool operator== (CoinPile & , CoinPile &);

	// for database creation:
	static char* giveAllCoinCountriesAsTextTupels(char* thirdcolumn);
	static char* giveAllCoinValuesAsTextTripels(char* fourthcolumn);
};


/////////////////////////////////////////////////////////////////////
// class Session
//
// can hold one purse (a list of different CoinPiles)
// with all the associated data
// like when, how long, how difficult, version
// plus the Usage-data for this session
/////////////////////////////////////////////////////////////////////

#include <qdatetime.h>
#include "usage.h"

class Session {
public:
	QValueList<CoinPile>	Purse;			// the coins
	QValueList<UserEvent>	SessionEvents;  // the UserEvents
	Q_UINT32 				SumCents;			// as a primitive checksum
	QDateTime				EntryDateTime;
	Q_UINT16				Version;

public:
    Session();
	Session(Q_UINT32 sumCents_, QDateTime datetime, Q_UINT16 progversion=PROGRAMVERSION);
	Session(const Session& Source);
	Session& operator=(Session& Source);

	void insertCoins(int coin, int country, Q_UINT32 number);
	void insertCoins(CoinPile cp);
	void emptyPurse();
	friend QDataStream& operator<< (QDataStream &s, Session SI);
	friend QDataStream& operator>> (QDataStream &s, Session &SI);

	// for exchange with CountTable:
	vector<vector<Q_UINT32> > PurseToCounttableArray();
	void SetPurseFromCounttableArray(const vector<vector<Q_UINT32> > &array);

	// Show data in Listview (server)
	void CreateListViewItemChildrenOf(QListViewItem *LV, Usage & Observer);

	// testing:
	friend bool operator== (Session & , Session &);
	friend vector<vector<Q_UINT32> > test3StorageClass(const vector<vector<Q_UINT32> > &CountTableArray);
};


/////////////////////////////////////////////////////////////////////
// class ClientStorage
//
// Holds all the Purses that were entried since last upload
// and all the other user information
/////////////////////////////////////////////////////////////////////

#include<qlist.h>
#include<qlistview.h>
#include<qdatastream.h>

#include "invitation.h"

class ClientStorage {
public:
	ClientStorage();
    ClientStorage(const ClientStorage&);
	void setDefaultValues();
	void CreateListViewItemChildrenOf(QListViewItem *LV, Usage &Observer);
	void MergeWithNewSessions(const ClientStorage & S);
	bool NecessarySettingsOK();
	void setNecessarySettingsOK(bool ok);
	int saveToDisk(QString filename);
	int loadFromDisk(QString filename);
	friend Q_UINT32 sizeOf(ClientStorage & Data);
	Q_UINT32 totalSumOfCents();

	friend QTextStream& operator<< (QTextStream &s, ClientStorage S);
	friend QDataStream& operator<< (QDataStream &s, ClientStorage S);
	friend QDataStream& operator>> (QDataStream &s, ClientStorage &S);
	friend void toStream  (QDataStream &s, QList<Session> &);
	friend void toStream  (QTextStream &s, QList<Session> &);
	friend void InitializeFromStream(QDataStream &s, QList<Session> &);

	// testing:
	friend void test4StorageClassSAVE(vector<vector<Q_UINT32> > array);
	friend vector<vector<Q_UINT32> > test4StorageClassLOAD();
	friend void test5StorageClass();

public:
	// program data:
	Q_UINT16 Version; 
	// Personified data:
	Q_UINT32 Membernumber;
	string Welcome; // WelcomeMessage
	QValueList<Q_UINT32> RecommenderMembernumbers;
	QString RecommenderName; 
	Q_UINT8 Country;         
	QString PostalCode;
	Q_UINT8 CityPopulation;  
	string eMailAddress;
	Q_UINT8 eMailFlags;		// 1= monthly 2=bimonthly 4=weekly ReminderEmails
							// 8 = WantsReminderEMails
							// 16= WantsRegularInformation on project
							// 32=WantsToHelp
							// 64=WantsToAnalyze
	Q_UINT16 SettingsFlags;   // 1=ServiceTabOn 
							  // 2=OpinionTabOn
	                          // 4=NecessarySettingsFilledIn
	                          // 8=OperatingSystem: 0=Windows, 1=Other (_WIN32 not defined)
	                          // 16=HasSeenSettings
							  // 32= Don't show "download pictures" at start
							  // 64= signed contract
							  // 128= allows observation
							  // 256= settings might have changed
	                          // 512= repaired missing-Greece membercountry


	//these are for user reminders
	QDateTime MemberSince;
	Q_UINT32 NumberOfUploads;  
	QDateTime LastUpload;
	Q_UINT16 NumberOfStartsSinceLastUpload;
	// The Number of the favourite Coin:
	Q_UINT8 FavouriteCoin;
	// the user can leave his/her opinion
	Q_UINT8 GuessWhenMixdateYear, GuessWhenMixdateMonth;
	Q_UINT8 GuessWhenArrivaldateYear, GuessWhenArrivaldateMonth;
	QString Message;    
	// this can be changed if server is moved

	string ServerAddress;
	Q_UINT16 ServerPort;
	string FTPServer; 
	// misc
	Q_UINT8 NumberOfPicturesAvailable;
	// Data Analysis read when?
	Q_UINT16 LastAnalysisPage;
	Q_UINT16 LastTxtMessage;
	Q_UINT16 LastPicture;  
	QDateTime LastDataAnalysisRead;
	// Window Geometry:
    QPoint WPosition;
    QSize  WSize;
	// user observation:
	QMap<Q_UINT16, Q_UINT32> TotalEvents;
	// all the stored purses:
	Q_UINT16				NumFullPurses;
	QValueList<Session> Sessions; 

	QValueList<InvitedPerson > InvitationList;
};


/////////////////////////////////////////////////////////////////////
// class ServerAnswer
//
/////////////////////////////////////////////////////////////////////
#include <qtextview.h>
#include "dialogs.h"

class ServerAnswer {
public:
	ServerAnswer ();
    ServerAnswer (const ServerAnswer&);
	bool ProcessAnswerAndStoreInto(ClientStorage * theData, Dialogs * theDialogs, 
									QTextView *infoText);

	friend Q_UINT32 sizeOf( ServerAnswer &);
	friend QDataStream & operator<< (QDataStream& s, ServerAnswer SA);
	friend QTextStream & operator<< (QTextStream& s, ServerAnswer SA);
	friend QDataStream & operator>> (QDataStream& s, ServerAnswer & SA);

	Q_UINT16		Version;    
	Q_UINT8			AnswerFlags;    // 1 ok
									// 2 AddressUpdateNecessary
									// 4 ProgramUpdateRecommended
									// 8 ProgramUpdateNecessary
									// 16 NewMember
									// 32 AnalysisDataAvailable

	string			ServerAddress;  // 
	Q_UINT16		ServerPort;		// if AddressUpdateNecessary
	string			FTPServer;		// 
	string			DownloadServer; // if ProgramUpdateNecessary

	Q_UINT32		NewMemberNumber; //
	string			NewMemberHello;  // if NewMember
	QDateTime		MemberSince;     //

	Q_UINT16		NewVersion;
	Q_UINT16		NumberOfNewestText; //
	Q_UINT16		NumberOfNewestPic;  // if AnalysisDataAvailable
	Q_UINT16		NumberOfNewestPage; // 

	QDate			NextSessionRecommendation;

	Q_UINT32		NumberOfTotalMembers;  
	Q_UINT32		NumberOfTotalPursesOnServer;
	Q_UINT32		TotalSumCents;
	Q_UINT8			LastUploadWasFromCountry;

	string			DailyMessage; // short message - or nothing
	string			MemberSpecificMessage; // to be able to respond to Message
};


#endif // STORAGE_H
